
import asyncio
import logging
from typing import Optional

from async_lru import alru_cache
from brownie import ZERO_ADDRESS, chain
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20
from y.constants import weth
from y.contracts import Contract
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils.dank_mids import dank_w3

logger = logging.getLogger(__name__)

if chain.id == 1:
    router = Contract("0xbAF9A5d4b0052359326A6CDAb54BABAa3a3A9643")
    gas_coin = weth
elif chain.id == 56:
    router = Contract("0xD41B24bbA51fAc0E4827b6F94C0D6DDeB183cD64")
    gas_coin = Contract("0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c") #wbnb
else: 
    router = None
    gas_coin = None

#yLazyLogger(logger)
def is_mooniswap_pool(token: AnyAddressType) -> bool:
    return await_awaitable(is_mooniswap_pool_async(token))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_mooniswap_pool_async(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    if router is None:
        return False
    return await router.isPool.coroutine(address)

def get_pool_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_pool_price_async(token, block=block))

#yLazyLogger(logger)
async def get_pool_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    token = await Contract.coroutine(address)
    token0, token1 = await asyncio.gather(
        token.token0.coroutine(),
        token.token1.coroutine(),
    )

    bal0, bal1, price0, price1, total_supply = await asyncio.gather(
        dank_w3.eth.get_balance(token.address) if token0 == ZERO_ADDRESS else ERC20(token0).balance_of_readable_async(token.address, block),
        dank_w3.eth.get_balance(token.address) if token1 == ZERO_ADDRESS else ERC20(token1).balance_of_readable_async(token.address, block),
        magic.get_price_async(gas_coin, block) if token0 == ZERO_ADDRESS else magic.get_price_async(token0, block),
        magic.get_price_async(gas_coin, block) if token1 == ZERO_ADDRESS else magic.get_price_async(token1, block),
        ERC20(address).total_supply_readable_async(block),
    )

    if token0 == ZERO_ADDRESS:
        bal0 /= 1e18
    elif token1 == ZERO_ADDRESS:
        bal1 /= 1e18

    totalVal = bal0 * price0 + bal1 * price1
    price = totalVal / total_supply
    return price
    



import logging
from collections import defaultdict
from itertools import count, product
from operator import itemgetter
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

import brownie
import requests
from brownie import chain, web3
from eth_abi.exceptions import InsufficientDataBytes
from web3.exceptions import CannotHandleRequest
from y import convert
from y.contracts import Contract, contract_creation_block
from y.datatypes import Address, AddressOrContract, AnyAddressType, Block
from y.exceptions import continue_if_call_reverted
from y.interfaces.multicall2 import MULTICALL2_ABI
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import _decimals, _totalSupply

from multicall import Call
from multicall.utils import await_awaitable, gather

logger = logging.getLogger(__name__)

MULTICALL2 = {
    Network.Mainnet:            "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
    Network.Arbitrum:           "0x842eC2c7D803033Edf55E478F461FC547Bc54EB2",
    Network.Avalanche:          "0xdf2122931FEb939FB8Cf4e67Ea752D1125e18858",
    Network.BinanceSmartChain:  "0xfF6FD90A470Aaa0c1B8A54681746b07AcdFedc9B",
    Network.Fantom:             "0xBAD2B082e2212DE4B065F636CA4e5e0717623d18",
    Network.Harmony:            "0x34b415f4d3b332515e66f70595ace1dcf36254c5",
    Network.Heco:               "0xd1F3BE686D64e1EA33fcF64980b65847aA43D79C",
    Network.Moonriver:          "0xaeF00A0Cf402D9DEdd54092D9cA179Be6F9E5cE3",
    Network.Polygon:            "0xc8E51042792d7405184DfCa245F2d27B94D013b6",
    Network.xDai:               "0x9903f30c1469d8A2f415D4E8184C93BD26992573",
    Network.Aurora:             "0xe0e3887b158F7F9c80c835a61ED809389BC08d1b",
    Network.Cronos:             "0x5e954f5972EC6BFc7dECd75779F10d848230345F",
}.get(chain.id, None)

multicall = None
multicall2 = brownie.Contract.from_abi("Multicall2",MULTICALL2, MULTICALL2_ABI) if chain.id in [Network.Harmony,Network.Cronos] else Contract(MULTICALL2)

# the address doesn't matter, it just needs to have code
code = " "

multicall_deploy_block = contract_creation_block(multicall2.address)


#yLazyLogger(logger)
def multicall_same_func_no_input(
    addresses: Iterable[AnyAddressType],
    method: str, 
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:

    return await_awaitable(
        multicall_same_func_no_input_async(addresses, method, block=block, apply_func=apply_func, return_None_on_failure=return_None_on_failure)
    )
    

#yLazyLogger(logger)
async def multicall_same_func_no_input_async(
    addresses: Iterable[AnyAddressType],
    method: str, 
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:

    addresses = _clean_addresses(addresses)
    results = await gather([Call(address, [method], [[address,apply_func]], block_id=block).coroutine() for address in addresses])
    return [v for call in results for k, v in call.items()]


#yLazyLogger(logger)
def multicall_same_func_same_contract_different_inputs(
    address: AnyAddressType, 
    method: str, 
    inputs: Union[List, Tuple],  
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:

    return await_awaitable(
        multicall_same_func_same_contract_different_inputs_async(address, method, inputs, block=block, apply_func=apply_func, return_None_on_failure=return_None_on_failure)
    )


#yLazyLogger(logger)
async def multicall_same_func_same_contract_different_inputs_async(
    address: AnyAddressType, 
    method: str, 
    inputs: Union[List, Tuple],  
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:
    assert inputs
    address = convert.to_address(address)
    results = await gather(Call(address, [method, input], [[input,apply_func]], block_id=block).coroutine() for input in inputs)
    return [result for call in results for key, result in call.items()]


#yLazyLogger(logger)
def multicall_decimals(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:
    return await_awaitable(
        multicall_decimals_async(addresses, block=block, return_None_on_failure=return_None_on_failure)
    )

#yLazyLogger(logger)
async def multicall_decimals_async(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:

    try: 
        return await gather([Call(str(address), ['decimals()(uint256)'], block_id=block).coroutine() for address in addresses])
    except (CannotHandleRequest,InsufficientDataBytes):
        pass # TODO investigate these
    except Exception as e:
        continue_if_call_reverted(e)

    decimals = await gather(
        [_decimals(address,block=block,return_None_on_failure=return_None_on_failure) for address in addresses]
    )
    return decimals


#yLazyLogger(logger)
def multicall_totalSupply(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:
    return await_awaitable(
        multicall_totalSupply_async(addresses, block=block, return_None_on_failure=return_None_on_failure)
    )

#yLazyLogger(logger)
async def multicall_totalSupply_async(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:

    try:
        return await multicall_same_func_no_input_async(addresses, 'totalSupply()(uint256)', block=block)
    except (CannotHandleRequest,InsufficientDataBytes):
        pass
        
    return [await _totalSupply(address,block=block,return_None_on_failure=return_None_on_failure) for address in addresses] 


#yLazyLogger(logger)
def fetch_multicall(*calls: Any, block: Optional[Block] = None) -> List[Optional[Any]]:
    # https://github.com/makerdao/multicall
    multicall_input = []
    fn_list = []
    decoded = []

    for contract, fn_name, *fn_inputs in calls:
        fn = getattr(contract, fn_name)

        # check that there aren't multiple functions with the same name
        if hasattr(fn, "_get_fn_from_args"):
            fn = fn._get_fn_from_args(fn_inputs)

        fn_list.append(fn)
        multicall_input.append((contract, fn.encode_input(*fn_inputs)))

    if isinstance(block, int) and block < multicall_deploy_block:
        # use state override to resurrect the contract prior to deployment
        data = multicall2.tryAggregate.encode_input(False, multicall_input)
        call = web3.eth.call(
            {'to': str(multicall2), 'data': data},
            block or 'latest',
            {str(multicall2): {'code': f'0x{multicall2.bytecode}'}},
        )
        result = multicall2.tryAggregate.decode_output(call)
    else:
        result = multicall2.tryAggregate.call(
            False, multicall_input, block_identifier=block or 'latest'
        )

    for fn, (ok, data) in zip(fn_list, result):
        try:
            assert ok, "call failed"
            decoded.append(fn.decode_output(data))
        except (AssertionError, InsufficientDataBytes):
            decoded.append(None)

    return decoded


#yLazyLogger(logger)
def multicall_matrix(contracts, params, block="latest"):
    matrix = list(product(contracts, params))
    calls = [[contract, param] for contract, param in matrix]

    results = fetch_multicall(*calls, block=block)

    output = defaultdict(dict)
    for (contract, param), value in zip(matrix, results):
        output[contract][param] = value

    return dict(output)


#yLazyLogger(logger)
def batch_call(calls: Tuple[Contract,str,List[Any],Block]) -> List[Any]:
    """
    Similar interface but block height as last param. Uses JSON-RPC batch.
    [[contract, 'func', arg, block_identifier]]
    """
    jsonrpc_batch = []
    fn_list = []
    ids = count()

    for contract, fn_name, *fn_inputs, block in calls:
        fn = getattr(contract, fn_name)
        if hasattr(fn, "_get_fn_from_args"):
            fn = fn._get_fn_from_args(fn_inputs)
        fn_list.append(fn)

        jsonrpc_batch.append(
            {
                'jsonrpc': '2.0',
                'id': next(ids),
                'method': 'eth_call',
                'params': [
                    {'to': str(contract), 'data': fn.encode_input(*fn_inputs)},
                    block,
                ],
            }
        )

    response = requests.post(web3.provider.endpoint_uri, json=jsonrpc_batch).json()
    return [
        fn.decode_output(res['result'])
        for res in sorted(response, key=itemgetter('id'))
    ]


#yLazyLogger(logger)
def _clean_addresses(
    addresses: Iterable[AnyAddressType]
    ) -> List[Address]:

    return [convert.to_address(address) for address in addresses]



import logging
from typing import Any, Callable, Optional, Union

import brownie
from async_lru import alru_cache
from brownie import ZERO_ADDRESS, convert, web3
from brownie.convert.datatypes import EthAddress
from eth_utils import encode_hex
from eth_utils import function_signature_to_4byte_selector as fourbyte
from y.contracts import Contract, proxy_implementation
from y.datatypes import Address, AddressOrContract, Block
from y.exceptions import (CalldataPreparationError, ContractNotVerified,
                          NonStandardERC20, NoProxyImplementation,
                          call_reverted)
from y.networks import Network
from y.utils.cache import memory
from y.utils.dank_mids import dank_w3
from y.utils.logging import yLazyLogger

from multicall.utils import await_awaitable, gather

logger = logging.getLogger(__name__)

"""
We use raw calls for commonly used functions because its much faster than using brownie Contracts
"""

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def _cached_call_fn(
    func: Callable,
    contract_address: AddressOrContract, 
    # Only supports one required arg besides contract_address for now
    block: Optional[Block], 
    required_arg = None,
    ) -> Any:
    if required_arg is None:
        return await func(contract_address, block=block)
    else:
        return await func(contract_address, required_arg, block=block)


#yLazyLogger(logger)
async def decimals(
    contract_address: AddressOrContract, 
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ) -> int:
    if block is None or return_None_on_failure is True: 
        return await _decimals(contract_address, block=block, return_None_on_failure=return_None_on_failure)
    else:
        return await _cached_call_fn(_decimals,contract_address,block)


#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def _decimals(
    contract_address: AddressOrContract, 
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ) -> Optional[int]:

    decimals = None

    # method 1
    # NOTE: this will almost always work, you will rarely proceed to further methods
    try: decimals = await raw_call_async(contract_address, "decimals()", block=block, output='int', return_None_on_failure=True)
    except OverflowError:
        # OverflowError means the call didn't revert, but we can't decode it correctly
        # the contract might not comply with standards, so we can possibly fetch 
        # using the verified non standard abi as a fallback
        try: 
            contract = await Contract.coroutine(contract_address)
            decimals = await contract.decimals.coroutine(block_identifier=block)
        except ContractNotVerified:
            pass
        except AttributeError as e:
            if "has no attribute 'decimals'" not in str(e):
                raise
            # we got a response from the chain but brownie can't find `DECIMALS` method, 
            # maybe our cached contract definition is messed up. let's repull it
            decimals = brownie.Contract.from_explorer(contract_address).decimals(block_identifier=block)

    if decimals is not None:
        return decimals

    # method 2
    try: decimals = await raw_call_async(contract_address, "DECIMALS()", block=block, output='int', return_None_on_failure=True)
    except OverflowError: 
        # OverflowError means the call didn't revert, but we can't decode it correctly
        # the contract might not comply with standards, so we can possibly fetch DECIMALS 
        # using the verified non standard abi as a fallback
        try: 
            contract = await Contract.coroutine(contract_address)
            decimals = await contract.DECIMALS.coroutine(block_identifier=block)
        except ContractNotVerified:
            pass
        except AttributeError as e:
            if "has no attribute 'DECIMALS'" not in str(e):
                raise
            # we got a response from the chain but brownie can't find `DECIMALS` method, 
            # maybe our cached contract definition is messed up. let's repull it
            decimals = brownie.Contract.from_explorer(contract_address).DECIMALS(block_identifier=block)
                
    if decimals is not None:
        return decimals

    # method 3
    try: decimals = await raw_call_async(contract_address, "getDecimals()", block=block, output='int', return_None_on_failure=True)
    except OverflowError: 
        # OverflowError means the call didn't revert, but we can't decode it correctly
        # the contract might not comply with standards, so we can possibly fetch DECIMALS 
        # using the verified non standard abi as a fallback
        try: 
            contract = await Contract.coroutine(contract_address)
            decimals = await contract.getDecimals.coroutine(block_identifier=block)
        except ContractNotVerified:
            pass
        except AttributeError as e:
            if "has no attribute 'getDecimals'" not in str(e):
                raise
            # we got a response from the chain but brownie can't find `DECIMALS` method, 
            # maybe our cached contract definition is messed up. let's repull it
            decimals = brownie.Contract.from_explorer(contract_address).getDecimals(block_identifier=block)
                
    if decimals is not None:
        return decimals

    # we've failed to fetch
    if return_None_on_failure:
        return None

    if await proxy_implementation(contract_address, block) == ZERO_ADDRESS:
        raise NoProxyImplementation(f"""
            Contract {contract_address} is a proxy contract, and had no implementation at block {block}.""")

    raise NonStandardERC20(f'''
        Unable to fetch `decimals` for {contract_address} on {Network.printable()}
        If the contract is verified, please check to see if it has a strangely named
        `decimals` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
        with the contract address and correct method name so we can keep things going smoothly :)''')


#yLazyLogger(logger)
async def _totalSupply(
    contract_address: AddressOrContract, 
    block: Optional[Block] = None,
    return_None_on_failure: bool = False # TODO: implement this kwarg
    ) -> Optional[int]:

    try: total_supply = await raw_call_async(contract_address, "totalSupply()", block=block, output='int', return_None_on_failure=True)
    except OverflowError:
        # OverflowError means the call didn't revert, but we can't decode it correctly
        # the contract might not comply with standards, so we can possibly fetch totalSupply 
        # using the verified non standard abi as a fallback
        try:
            contract = await Contract.coroutine(contract_address)
            total_supply = await contract.totalSupply.coroutine(block_identifier=block)
        except AttributeError as e:
            if "has no attribute 'totalSupply'" not in str(e):
                raise
            # we got a response from the chain but brownie can't find `totalSupply` method, 
            # maybe our cached contract definition is messed up. let's repull it
            total_supply = brownie.Contract.from_explorer(contract_address).totalSupply(block_identifier=block)

    if total_supply is not None:
        return total_supply

    if return_None_on_failure:
        return None
    raise NonStandardERC20(f'''
        Unable to fetch `totalSupply` for {contract_address} on {Network.printable()}
        If the contract is verified, please check to see if it has a strangely named
        `totalSupply` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
        with the contract address and correct method name so we can keep things going smoothly :)''')


#yLazyLogger(logger)
async def _totalSupplyReadable(
    contract_address: AddressOrContract, 
    block: Optional[Block] = None,
    return_None_on_failure: bool = False
    ) -> Optional[float]:

    total_supply = _totalSupply(contract_address, block=block, return_None_on_failure=return_None_on_failure)
    decimals = _decimals(contract_address, block=block, return_None_on_failure=return_None_on_failure)
    total_supply, decimals = await gather([total_supply, decimals])

    if total_supply is not None and decimals is not None:
        return total_supply / 10 ** decimals

    if total_supply is None and decimals is None and return_None_on_failure:
        return None

    raise NonStandardERC20(f'''
        Unable to fetch `totalSupplyReadable` for {contract_address} on {Network.printable()}
        totalSupply: {total_supply} decimals: {decimals}
        If the contract is verified, please check to see if it has a strangely named `totalSupply` or
        `decimals` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
        with the contract address and correct function name so we can keep things going smoothly :)''')

def balanceOf(
    contract_address: AddressOrContract, 
    address: Address, 
    block: Optional[Block] = None,
    return_None_on_failure: bool = False
    ) -> Optional[int]:
    return await_awaitable(balanceOf_async(contract_address, address, block=block, return_None_on_failure=return_None_on_failure))

#yLazyLogger(logger)
async def balanceOf_async(
    call_address: AddressOrContract, 
    input_address: AddressOrContract, 
    block: Optional[Block] = None,
    return_None_on_failure: bool = False
    ) -> Optional[int]:

    # method 1
    # NOTE: this will almost always work, you will rarely proceed to further methods
    try:
        balance = await raw_call_async(call_address, "balanceOf(address)", block=block, inputs=input_address, output='int', return_None_on_failure=True)
    except OverflowError:
        # OverflowError means the call didn't revert, but we can't decode it correctly
        # the contract might not comply with standards, so we can possibly fetch balanceOf 
        # using the verified non standard abi as a fallback
        try:
            contract = await Contract.coroutine(call_address)
            return await contract.balanceOf.coroutine(input_address, block_identifier=block)
        except AttributeError as e:
            if "has no attribute 'balanceOf'" not in str(e):
                raise
            # we got a response from the chain but brownie can't find `balanceOf` method, 
            # maybe our cached contract definition is messed up. let's repull it
            balance = brownie.Contract.from_explorer(call_address).balanceOf(input_address, block_identifier=block)
    
    if balance is not None:
        return balance

    # we've failed to fetch
    if return_None_on_failure:
        return None
    raise NonStandardERC20(f'''
        Unable to fetch `balanceOf` for token: {call_address} holder: {input_address} on {Network.printable()}
        If the contract is verified, please check to see if it has a strangely named
        `totalSupply` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
        with the contract address and correct function name so we can keep things going smoothly :)''')


#yLazyLogger(logger)
async def _balanceOfReadable(
    call_address: AddressOrContract, 
    input_address: AddressOrContract, 
    block: Optional[Block] = None,
    return_None_on_failure: bool = False
    ) -> Optional[float]:

    # TODO _balanceOf return_None_on_failure
    balance = balanceOf_async(call_address, input_address, block=block, return_None_on_failure=return_None_on_failure)
    decimals = _decimals(call_address, block=block, return_None_on_failure=return_None_on_failure)
    balance, decimals = await gather([balance, decimals])
    if balance is not None and decimals is not None:
        return balance / 10 ** decimals

    # we've failed to fetch
    if return_None_on_failure:
        return None
    else:
        raise NonStandardERC20(
        f'''Unable to fetch `balanceOfReadable` for
        token: {call_address} holder: {input_address} on {Network.printable()}
        balanceOf: {balance} decimals: {decimals}
        If the contract is verified, please check to see if it has a strangely named `balanceOf` or
        `decimals` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
        with the contract address and correct function name so we can keep things going smoothly :)''')


#yLazyLogger(logger)
def raw_call(
    contract_address: AddressOrContract, 
    method: str, 
    block: Optional[Block] = None, 
    inputs = None, 
    output: str = None,
    return_None_on_failure: bool = False
    ) -> Optional[Any]:

    if type(contract_address) != str:
        contract_address = str(contract_address)

    data = {'to': convert.to_address(contract_address),'data': prepare_data(method,inputs)}

    try: response = web3.eth.call(data,block_identifier=block)
    except ValueError as e:
        if return_None_on_failure is False:                 raise
        if call_reverted(e):                                return None
        if 'invalid opcode' in str(e):                      return None
        else:                                               raise
    
    if output is None:                                      return response
    elif output == 'address' and response.hex() == '0x':    return ZERO_ADDRESS
    elif output == 'address':                               return convert.to_address(f'0x{response.hex()[-40:]}')
    elif output in [int, 'int','uint','uint256']:           return convert.to_int(response)
    elif output in [str, 'str']:                            return convert.to_string(response).replace('\x00','').strip()

    else: raise TypeError('Invalid output type, please select from ["str","int","address"]')


#yLazyLogger(logger)
async def raw_call_async(
    contract_address: AddressOrContract, 
    method: str, 
    block: Optional[Block] = None, 
    inputs = None, 
    output: str = None,
    return_None_on_failure: bool = False
    ) -> Optional[Any]:
    '''
    call a contract with only address and method. Bypasses brownie Contract object formation to save time
    only works with 1 input, ie `balanceOf(address)` or `getPoolInfo(uint256)`
    '''

    if type(contract_address) != str:
        contract_address = str(contract_address)

    data = {'to': convert.to_address(contract_address),'data': prepare_data(method,inputs)}

    try: response = await dank_w3.eth.call(data,block_identifier=block)
    except ValueError as e:
        if return_None_on_failure is False:                 raise
        if call_reverted(e):                                return None
        if 'invalid opcode' in str(e):                      return None
        else:                                               raise
    
    if output is None:                                      return response
    elif output == 'address' and response.hex() == '0x':    return ZERO_ADDRESS
    elif output == 'address':                               return convert.to_address(f'0x{response.hex()[-40:]}')
    elif output in [int, 'int','uint','uint256']:           return convert.to_int(response)
    elif output in [str, 'str']:                            return convert.to_string(response).replace('\x00','').strip()

    else: raise TypeError('Invalid output type, please select from ["str","int","address"]')


#yLazyLogger(logger)
def prepare_data(
    method, 
    inputs = Union[None, bytes, int, str, Address, EthAddress, brownie.Contract, Contract]
    ) -> str:
    method = encode_hex(fourbyte(method))

    if inputs is None:
        return method

    elif type(inputs) in [bytes, int, str, Address, EthAddress, brownie.Contract, Contract]:
        return method + prepare_input(inputs)
    
    raise CalldataPreparationError(f'''
        Supported types are: Union[None, bytes, int, str, Address, EthAddress, brownie.Contract, y.Contract]
        You passed {type(inputs)} {inputs}''')
    
    # these don't work yet, wip
    '''
    elif type(inputs) in (Set,List,Tuple):
        # this doesn't work right now
        inputs = str(concat([method,*[prepare_input(_) for _ in inputs]]))
        logger.warning(inputs)
        return encode_hex(inputs)
    elif type(inputs) == Dict:
        # this doesn't work right now
        inputs = str(concat([method,*[prepare_input(_) for _ in inputs.values()]]))
        logger.warning(inputs)
        return encode_hex(inputs)
    '''


#yLazyLogger(logger)
def prepare_input(
    input: Union[
        bytes, # for bytes input
        int, # for int input
        Union[str, Address, EthAddress, brownie.Contract, Contract] # for address input
        ]
    ) -> str:
    
    input_type = type(input)

    if input_type == bytes:
        return input.hex()

    if input_type == int:
        return convert.to_bytes(input).hex()

    # we can't process actual strings so we can assume that any string input is an address. regular strings will fail
    if input_type in [str, Address, EthAddress, brownie.Contract, Contract]:
        return '000000000000000000000000' + convert.to_address(input)[2:]
    
    raise CalldataPreparationError(f'''
        Supported input types are
        uint: int, 
        address: Union[str, Address, brownie.Contract, y.Contract]
        you passed input: {input} type: {input_type}
        ''')




import logging
from typing import Optional

from async_lru import alru_cache
from multicall.utils import await_awaitable, gather
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods_async
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)

#yLazyLogger(logger)
def is_eps_rewards_pool(token_address: AnyAddressType) -> bool:
    return await_awaitable(is_eps_rewards_pool_async(token_address))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_eps_rewards_pool_async(token_address: AnyAddressType) -> bool:
    return await has_methods_async(token_address, ('lpStaker()(address)','rewardTokens(uint)(address)','rewardPerToken(address)(uint)','minter()(address)'))

#yLazyLogger(logger)
def get_price(token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_async(token_address, block))

async def get_price_async(token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
    minter = await raw_call_async(token_address,'minter()',output='address', block=block)
    minter = await Contract.coroutine(minter)
    i, balances = 0, []
    while True:
        try:
            coin, balance = await gather([
                minter.coins.coroutine(i, block_identifier = block),
                minter.balances(i, block_identifier = block),
            ])
            balance /= await ERC20(coin).scale
            balances.append(WeiBalance(balance, coin, block))
            i += 1
        except:
            break
    coin_values, total_supply = await gather([
        gather([b.value_usd_async for b in balances]),
        ERC20(token_address).total_supply_readable_async(block),
    ])
    tvl = sum(coin_values)
    return UsdPrice(tvl / total_supply)



import logging
from functools import lru_cache

from async_lru import alru_cache
from multicall.utils import await_awaitable
from y import convert
from y.constants import STABLECOINS
from y.datatypes import AnyAddressType
from y.prices import convex, one_to_one, popsicle, yearn
from y.prices.chainlink import chainlink
from y.prices.dex import mooniswap
from y.prices.dex.balancer import balancer_multiplexer
from y.prices.dex.genericamm import generic_amm
from y.prices.dex.uniswap import uniswap_multiplexer
from y.prices.eth_derivs import creth, wsteth
from y.prices.lending import ib
from y.prices.lending.aave import aave
from y.prices.lending.compound import compound
from y.prices.stable_swap import (belt, ellipsis, froyo, mstablefeederpool,
                                  saddle)
from y.prices.stable_swap.curve import curve
from y.prices.synthetix import synthetix
from y.prices.tokenized_fund import basketdao, gelato, piedao, tokensets
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

@lru_cache(maxsize=None)
def check_bucket(
    token: AnyAddressType
    ) -> str:
    return await_awaitable(check_bucket_async(token))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def check_bucket_async(
    token: AnyAddressType
    ) -> str:

    token_address = convert.to_address(token)

    # these require neither calls to the chain nor contract initialization
    if token_address == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE":       return 'wrapped gas coin'
    elif token_address in STABLECOINS:                                      return 'stable usd'
    elif one_to_one.is_one_to_one_token(token_address):                     return 'one to one'
    
    elif wsteth.is_wsteth(token_address):                                   return 'wsteth'
    elif creth.is_creth(token_address):                                     return 'creth'
    elif belt.is_belt_lp(token_address):                                    return 'belt lp'

    elif froyo.is_froyo(token_address):                                     return 'froyo'
    elif aave.is_atoken(token_address):                                     return 'atoken'
    elif convex.is_convex_lp(token_address):                                return 'convex'

    # these just require calls
    elif await balancer_multiplexer.is_balancer_pool_async(token_address):  return 'balancer pool'
    elif await ib.is_ib_token_async(token_address):                         return 'ib token'

    elif await gelato.is_gelato_pool_async(token_address):                  return 'gelato'
    elif await piedao.is_pie_async(token_address):                          return 'piedao lp'
    elif await tokensets.is_token_set_async(token_address):                 return 'token set'

    elif await ellipsis.is_eps_rewards_pool_async(token_address):           return 'ellipsis lp'
    elif await mstablefeederpool.is_mstable_feeder_pool(token_address):     return 'mstable feeder pool'
    elif await saddle.is_saddle_lp_async(token_address):                    return 'saddle'

    elif await basketdao.is_basketdao_index_async(token_address):           return 'basketdao'
    elif await popsicle.is_popsicle_lp_async(token_address):                return 'popsicle'


    # these require both calls and contract initializations
    elif await uniswap_multiplexer.is_uniswap_pool_async(token_address):    return 'uni or uni-like lp'
    # this just requires contract initialization but should go behind uniswap
    elif token_address in generic_amm:                                      return 'generic amm'
    elif await mooniswap.is_mooniswap_pool_async(token_address):            return 'mooniswap lp'
    elif await compound.is_compound_market_async(token_address):            return 'compound'
    elif token_address in curve:                                            return 'curve lp'
    elif chainlink and await chainlink.has_feed(token_address):                           return 'chainlink feed'
    elif synthetix and await synthetix.is_synth(token_address):             return 'synthetix'
    elif await yearn.is_yearn_vault_async(token_address):                   return 'yearn or yearn-like'
